/**
 * mm_options.cpp - implementation of MMOptions class for parsing mountmon options
 * Copyright (C) 2007 Chumby Industries. All rights reserved.
 * henry@chumby.com
**/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>

#include "mm_options.h"

// Valid environment variable name characters
static char *validEnvChars = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789_";

MMOptions::MMOptions()
{
	m_conf[0] = "/etc/mountmon/mountmon.conf";
	m_curInclude = 0;
	m_envCount = 0;
	m_verbose = -1;
	m_verboseRequested = 0;
}

MMOptions::~MMOptions()
{
}

// Set configuration file. Default is /etc/mountmon/mountmon.conf
void MMOptions::SetConf( const char *confPath )
{
	m_conf[m_curInclude] = confPath;
}

// Parse configuration. Return number of rules processed successfully.
int MMOptions::ParseConfig( int verbose )
{
	int rulesParsed = -1;
	m_verboseRequested = verbose;
	FILE *f = fopen( m_conf[m_curInclude].c_str(), "r" );
	if (f != NULL)
	{
		if (verbose) printf( "Parsing level %d, source=%s\n", m_curInclude, m_conf[m_curInclude].c_str() );
		rulesParsed = 0;
		char buff[1024];
		while (fgets( buff, sizeof(buff), f ))
		{
			char *start = buff + strspn( buff, " \t" );
			if (*start == '#' || *start == '\r' || *start == '\n' || *start == '\0')
			{
				continue;
			}
			char *nextToken = start + strcspn( start, " =\t" );
			nextToken += strspn( nextToken, " =\t" );
			char *nextEnd = nextToken + strcspn( nextToken, " \t\r\n" );
			if (verbose) printf( "Processing %s", start );
			if (!strncmp( start, "mount_cmd", 9 ))
			{
				*nextEnd = '\0';
				if (*nextToken)
				{
					rulesParsed++;
					m_mountCmd = nextToken;
					if (verbose) printf( "New mount cmd = %s\n", m_mountCmd.c_str() );
				}
				continue;
			}
			if (!strncmp( start, "umount_cmd", 10 ))
			{
				*nextEnd = '\0';
				if (*nextToken)
				{
					rulesParsed++;
					m_umountCmd = nextToken;
				}
				continue;
			}
			if (!strncmp( start, "add_cmd", 7 ))
			{
				*nextEnd = '\0';
				if (*nextToken)
				{
					rulesParsed++;
					m_addCmd = nextToken;
				}
				continue;
			}
			if (!strncmp( start, "remove_cmd", 10 ))
			{
				*nextEnd = '\0';
				if (*nextToken)
				{
					rulesParsed++;
					m_removeCmd = nextToken;
				}
				continue;
			}
			if (!strncmp( start, "verbose", 7 ))
			{
				*nextEnd = '\0';
				if (*nextToken)
				{
					rulesParsed++;
					m_verbose = atoi( nextToken );
					if (m_verbose >= 0)
					{
						verbose = m_verbose;
					}
				}
				continue;
			}
			if (!strncmp( start, "env", 3 ))
			{
				// We can have
				// env { key } = value
				// env [ key ] = value
				// Make sure braces match
				char *openBracket = strpbrk( start, "[{" );
				if (openBracket)
				{
					char *keyValue = openBracket + 1 + strspn( openBracket + 1, " \t" );
					char *closeBracket = strchr( start, *openBracket == '[' ? ']' : '}' );
					if (closeBracket)
					{
						char *setValue = closeBracket + 1 + strspn( closeBracket + 1, " \t" );
						if (*setValue == '=')
						{
							setValue++;
							setValue += strspn( setValue, " \t" );
						}
						while (closeBracket > openBracket && closeBracket[-1] == ' ')
						{
							closeBracket--;
						}
						*closeBracket = '\0';
						char *setValueEnd = setValue + strcspn( setValue, "\r\n" );
						*setValueEnd = '\0';
						AddEnvDef( keyValue, setValue );
					}
				}
				continue;
			}
			if (!strncmp( start, "include", 7 ))
			{
				*nextEnd = '\0';
				if (verbose) printf( "Processing include, new level %d\n", m_curInclude + 1 );
				// Are we out of levels?
				if (m_curInclude < MAX_INCLUDE_DEPTH)
				{
					// Have we already gotten this one?
					int n;
					bool isDuplicate = false;
					for (n = m_curInclude; n >= 0 && !isDuplicate; n--)
					{
						if (!strcmp( m_conf[n], nextToken ))
						{
							isDuplicate = true;
						}
					}
					if (!isDuplicate)
					{
						m_curInclude++;
						if (verbose) printf( "Processing %s\n", nextToken );
						m_conf[m_curInclude] = nextToken;
						int res = ParseConfig( verbose );
						if (res > 0)
						{
							if (verbose) printf( "Return from %s = %d\n", nextToken, res );
							rulesParsed += res;
						}
						else if (verbose)
						{
							if (verbose) printf( "Return from %s = %d (ERROR)\n", nextToken, res );
						}
					}
				}
				continue;
			}
			printf( "Unrecognized config line - ignoring: %s", start );
		}
		fclose( f );
	}
	return rulesParsed;
}

// Get include at specified level, where 0 is the top level (first file to be parsed)
const char *MMOptions::GetInclude( int depth ) const
{
	if (depth < 0 || depth > m_curInclude)
	{
		return NULL;
	}
	return m_conf[m_curInclude].c_str();
}

// Return index or -1 if not found
int MMOptions::FindEnvName( const char *name )
{
	int n;
	for (n = 0; n < m_envCount; n++)
	{
		if (m_envName[n] == name)
		{
			return n;
		}
	}
	// Not found
	return -1;
}

// Add name:value pair, true if successful false if failed (e.g. max exceeded)
bool MMOptions::AddEnvDef( const char *name, const char *value )
{
	int existing = FindEnvName( name );
	if (m_verboseRequested) printf( "AddEnvDef(%s,%s) find returned %d\n", name, value, existing );
	if (existing < 0)
	{
		// Add new entry if there's space
		if (m_envCount < MAX_ENV_CHANGES)
		{
			existing = m_envCount;
			if (m_verboseRequested) printf( "AddEnvDef() adding new value as %d\n", existing );
			m_envCount++;
			m_envName[existing] = name;
		}
		else
		{
			return false;
		}
	}
	m_envValue[existing] = value;
	// expand environment values
	int expandRes = ExpandEnvValues( m_envValue[existing], existing );
	if (m_verboseRequested) printf( "ExpandEnvValues() returned %d, new value [%s]\n", expandRes, m_envValue[existing].c_str() );
	return true;
}

int MMOptions::ExpandEnvValues( cstring& s, int exclude )
{
	int substitutions = 0;
	cstring r;
	const char *src = s.c_str();
	const char *srcEnd = src + s.length();
	if (m_verboseRequested) printf( "ExpandEnvValues(%s,%d) entry length=%d\n", s.c_str(), exclude, s.length() );
	while (src < srcEnd)
	{
		int normalSpan = strcspn( src, "\\$" );
		if (normalSpan)
		{
			if (m_verboseRequested > 1) printf( "Got normal span of %d chars\n", normalSpan );
			r.append( src, normalSpan );
			src += normalSpan;
			if (m_verboseRequested > 1) printf( "New r=%s\n", r.c_str() );
		}
		if (*src == '\\')
		{
			int escapedSpan = src[1] ? 2 : 1;
			r.append( src, escapedSpan );
			src += escapedSpan;
			// Go around again to catch another normal span
			continue;
		}
		if (m_verboseRequested > 2) printf( "src=%s", src );
		// We should have *src == '$'
		// Get end of variable - delimiters are whitespace and punctuation
		int varSpan;
		for (varSpan = 1;
			src+varSpan < srcEnd && strchr( validEnvChars, src[varSpan]) != NULL;
			varSpan++)
				;
		// If no length we have a trailing $
		if (varSpan == 1)
		{
			r.append( src, 1 );
			src++;
			continue;
		}
		// Get variable name without leading $
		cstring varName( src+1, varSpan-1 );
		if (m_verboseRequested) printf( "Got var name %s\n", varName.c_str() );
		// Look in config values
		int ourIndex = FindEnvName( varName.c_str() );
		// If excluded, fall through to getenv
		if (ourIndex >= 0 && ourIndex != exclude)
		{
			r += GetEnvValue( ourIndex );
			src += varSpan;
			substitutions++;
			continue;
		}
		// Check in environment
		char *envValue = getenv( varName.c_str() );
		// If undefined, treat it as empty
		if (envValue)
		{
			r += envValue;
			substitutions++;
		}
		src += varSpan;
	}
	if (substitutions)
	{
		s = r;
	}
	return substitutions;
}
