/**
mountmon - lightweight mount monitor daemon to allow scripting of udev events

20070518 henry@chumby.com Updated to use simple script model

The guts of this was taken from udevmonitor

**/
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdarg.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <getopt.h>
#include <syslog.h>
#include <pwd.h>
#include <time.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/select.h>
#include <linux/types.h>
#include <linux/netlink.h>

#include "udev.h"
#include "udevd.h"

#include "mm_options.h"

#define VER_STR "1.27"

static int uevent_netlink_sock = -1;
static int udev_monitor_sock = -1;
static int lfp = -1;
static volatile int udev_exit;
int kernel = 0;
int udev = 1;

// Close critical handles around scripts
int g_closeHandles = 0;

// Global debug level flag
int g_dbg = 0;

// Global daemon flag - 1 if we're daemonized
int g_daemon = 0;

// Print message to stdout or syslog depending on daemon status
void msgOut( int priority, const char *fmt, ...)
{
	va_list args;
	va_start( args, fmt );
	if (g_daemon)
	{
		vsyslog( priority, fmt, args );
	}
	else
	{
		vprintf( fmt, args );
	}
	va_end( args );
}

// Add variable to environment if not already listed
#define MAXENV 1024
char *g_envNames[MAXENV];
int g_envNameCount = 0;
void SaveEnvName( const char *name )
{
	int n;
	for (n = 0; n < g_envNameCount; n++)
	{
		if (!strcmp( g_envNames[n], name ))
		{
			return;
		}
	}
	// Add to end iff there is room
	if (g_envNameCount < MAXENV)
	{
		g_envNames[g_envNameCount] = strdup( name );
		g_envNameCount++;
	}
}

// Push variable to environment, to be released later with ReleaseEnv
void PushEnv( const char *name, const char *value )
{
	SaveEnvName( name );
	setenv( name, value, 1 );
}

// Release environment variables added via PushEnv
void ReleaseEnv()
{
	int n;
	for (n = 0; n < g_envNameCount; n++)
	{
		unsetenv( g_envNames[n] );
	}
}

// Map devpath to mount point such as /mnt/ipod, /mnt/fdisk, /mnt/cdrom, etc
typedef struct _autoMountPoint {
	cstring mountPoint;
	cstring usedBy;
	int status; // 0 = free, 1 = engaged, 2 = in use
} MountPoint_t;

// Map devpath to last mount point it used
typedef struct _lastUsage {
	cstring devPath;
	cstring mountPointUsed;
} LastUsage_t;

MountPoint_t g_mounts[128];
int g_mountsCount = 0;
#define MAXMOUNTS (sizeof(g_mounts)/sizeof(g_mounts[0]))

LastUsage_t g_lastUse[64];
int g_lastUseCount = 0;
#define MAXLASTUSE (sizeof(g_lastUse)/sizeof(g_lastUse[0]))

// Given a device node such as /dev/sda1, translate to a devpath
char *NodeToDevpath( const char *devnode )
{
	// Strip leading /dev
	if (!strncmp( devnode, "/dev", 4 ))
	{
		devnode += 4;
	}
	// Search /sys/block/*
	char buff[512];
	char cmd[512];
	snprintf( cmd, sizeof(cmd), "find /sys/block -name %s", *devnode=='/' ? (devnode + 1) : devnode );
	FILE *find = popen( cmd, "r" );
	if (g_dbg) msgOut( LOG_DEBUG, "executing '%s'\n", cmd );
	if (find)
	{
		char *devpath = NULL;
		while (fgets( buff, sizeof(buff), find ))
		{
			if (g_dbg) msgOut( LOG_DEBUG, "got %s", buff );
			devpath = strtok( &buff[4], "\n" ); // Skip /sys
			break;
		}
		pclose( find );
		if (devpath)
		{
			if (g_dbg) msgOut( LOG_DEBUG, "found devpath %s\n", devpath );
			return strdup( devpath );
		}
	}
	// Fallback
	snprintf( buff, sizeof(buff), "/block%s%s", *devnode=='/' ? "" : "/", devnode );
	if (g_dbg) msgOut( LOG_DEBUG, "fallback to %s\n", buff );
	return strdup( buff );
}

// Find mount point or add a new one. Return index or -1 if error (bogus mount point or out of nodes)
int FindMountPoint( const char *mountPoint )
{
	int n;
	// Reject all the bogus mount points like none, devfs, etc.
	if (!mountPoint || !*mountPoint || *mountPoint != '/')
	{
		return -1;
	}
	for (n = 0; n < g_mountsCount; n++)
	{
		if (!strcmp( g_mounts[n].mountPoint, mountPoint ))
		{
			return n;
		}
	}
	// Not found - add new entry if room
	if (g_mountsCount < MAXMOUNTS)
	{
		n = g_mountsCount;
		g_mountsCount++;
		g_mounts[n].status = 0;
		g_mounts[n].mountPoint = mountPoint;
		g_mounts[n].usedBy = "";
		return n;
	}
	return -1;
}

// Dump all mount points
void DumpMountPoints( int minStatus = 0 )
{
	int n;
	for (n = 0; n < g_mountsCount; n++)
	{
		if (g_mounts[n].status < minStatus)
		{
			continue;
		}
		msgOut( LOG_INFO, "mount[%d] %s usedby %s status %d\n", n, g_mounts[n].mountPoint.c_str(), g_mounts[n].usedBy.c_str(), g_mounts[n].status );
	}
}

// Return index + 1 of entry in g_mounts, or 0 if not mounted
int IsAlreadyMounted( const char *devpath )
{
	int n;
	for (n = 0; n < g_mountsCount; n++)
	{
		if (g_mounts[n].usedBy == devpath )
		{
			if (g_mounts[n].status > 0)
			{
				return n+1;
			}
		}
	}
	return 0;
}

// Return mount point where device is currently mounted or "" if not mounted
const char *GetCurrentMountPoint( const char *devpath )
{
	int mountIndex = IsAlreadyMounted( devpath );
	if (mountIndex <= 0)
	{
		return "";
	}
	return g_mounts[mountIndex-1].mountPoint.c_str();
}


// Set last mount point used by devpath
void UpdateMountHistory( const char *devpath, const char *mountPoint )
{
	int n;
	for (n = 0; n < g_lastUseCount; n++)
	{
		if (g_lastUse[n].devPath == devpath)
		{
			g_lastUse[n].mountPointUsed = mountPoint;
			return;
		}
	}
	// Add this if there's room
	if (g_lastUseCount < MAXLASTUSE)
	{
		g_lastUse[g_lastUseCount].devPath = devpath;
		g_lastUse[g_lastUseCount].mountPointUsed = mountPoint;
		g_lastUseCount++;
	}
}

// Get last mount point used or "" if not in history
const char *GetMountHistory( const char *devpath )
{
	int n;
	if (!devpath || !*devpath)
	{
		return "";
	}
	for (n = 0; n < g_lastUseCount; n++)
	{
		if (g_lastUse[n].devPath == devpath)
		{
			return g_lastUse[n].mountPointUsed.c_str();
		}
	}
	// No history
	return "";
}

// Mark mount point as in use
void MarkMountPoint( const char *devpath, const char *mountPoint, int newStatus = 2 )
{
	int n;
	for (n = 0; n < g_mountsCount; n++)
	{
		if (g_mounts[n].mountPoint == mountPoint )
		{
			g_mounts[n].usedBy = devpath;
			g_mounts[n].status = newStatus;
			UpdateMountHistory( devpath, mountPoint );
			return;
		}
	}
	// We need to add this mount point to list
	if (g_mountsCount >= MAXMOUNTS)
	{
		msgOut( LOG_ERR, "Error: cannot add mount point %s to list, too many entries!\n", mountPoint );
		return;
	}
	n = g_mountsCount;
	g_mountsCount++;
	g_mounts[n].usedBy = devpath;
	g_mounts[n].status = newStatus;
	g_mounts[n].mountPoint = mountPoint;
	UpdateMountHistory( devpath, mountPoint );
}

// Release mount point which we previously claimed
void ReleaseMountPoint( const char *devpath )
{
	int n;
	for (n = 0; n < g_mountsCount; n++)
	{
		if (g_mounts[n].usedBy == devpath )
		{
			msgOut( LOG_INFO, "<<< Releasing mount point %s formerly used by %s\n", g_mounts[n].mountPoint.c_str(), devpath );
			g_mounts[n].status = 0;
			return;
		}
	}
}

// Check a single mount point for a match
int CheckMount( const char *mountPoint, const char *devnode )
{
	char *devpath;
	// Is it one of our mount points?
	int mpIndex = FindMountPoint( mountPoint );
	if (mpIndex < 0)
	{
		return 0;
	}
	// Translate device node to devpath
	devpath = NodeToDevpath( devnode );
	// Mark as used
	// If not found use node name
	if (!devpath)
	{
		msgOut( LOG_ERR, "could not find devpath for device node %s using mount point %s\n", devnode, mountPoint );
		MarkMountPoint( devnode, mountPoint );
	}
	else
	{
		if (g_dbg) msgOut( LOG_DEBUG, "marking mount point %s as in use by devnode %s == devpath %s\n", mountPoint, devnode, devpath );
		MarkMountPoint( devpath, mountPoint );
		free( devpath );
	}
	return 1;
}

// Populate existing mounts already existing at startup
int CheckExistingMounts()
{
	// Get devnode mountpoint from /proc/mounts
	char buff[512];
	char *mountPoint, *devnode;
	int total = 0;
	FILE *mounts = fopen( "/proc/mounts", "r" );
	if (!mounts)
	{
		msgOut( LOG_ERR, "Error: cannot open /proc/mounts\n" );
		return -1;
	}
	while (fgets( buff, sizeof(buff), mounts ))
	{
		int found;
		devnode = strtok( buff, " \t\n" );
		mountPoint = strtok( NULL, " \t\n" );
		if (!mountPoint) continue;
		// Ignore devnodes that don't begin with /dev
		if (strncmp( devnode, "/dev", 4 )) continue;
		found = CheckMount( mountPoint, devnode );
		if (found > 0) total += found;
	}
	fclose( mounts );
	return total;
}


#define DAEMON_NAME "mountmon"
#define LOCKFILE_NAME "/var/run/" DAEMON_NAME ".pid"

/* Change this to the user under which to run */
#define RUN_AS_USER "root"

#define EXIT_SUCCESS 0
#define EXIT_FAILURE 1

static void child_handler(int signum)
{
    switch(signum) {
    case SIGALRM: exit(EXIT_FAILURE); break;
    case SIGUSR1: exit(EXIT_SUCCESS); break;
    case SIGCHLD: exit(EXIT_FAILURE); break;
    }
}

int OpenLockfile( const char *lockfile )
{
	lfp = open(lockfile,O_RDWR|O_CREAT,0640);
	if ( lfp < 0 ) {
		msgOut( LOG_ERR, "unable to create lock file %s, code=%d (%s)\n",
				lockfile, errno, strerror(errno) );
		return -1;
	}
	return lfp;
}

static void daemonize( const char *lockfile )
{
    pid_t pid, sid, parent;

    /* already a daemon */
    if ( getppid() == 1 ) return;

    /* Create the lock file as the current user */
    if ( lockfile && lockfile[0] ) {
    	if (OpenLockfile( lockfile ) < 0)
    	{
    		exit( EXIT_FAILURE );
    	}
    }

    /* Drop user if there is one, and we were run as root */
    if ( getuid() == 0 || geteuid() == 0 ) {
        struct passwd *pw = getpwnam(RUN_AS_USER);
        if ( pw ) {
            msgOut( LOG_NOTICE, "setting user to " RUN_AS_USER "\n" );
            setuid( pw->pw_uid );
        }
    }

    /* Trap signals that we expect to recieve */
    signal(SIGCHLD,child_handler);
    signal(SIGUSR1,child_handler);
    signal(SIGALRM,child_handler);

    /* Fork off the parent process */
    pid = fork();
    if (pid < 0) {
        msgOut( LOG_ERR, "unable to fork daemon, code=%d (%s)\n",
                errno, strerror(errno) );
        exit(EXIT_FAILURE);
    }
    /* If we got a good PID, then we can exit the parent process. */
    if (pid > 0) {

        /* Wait for confirmation from the child via SIGTERM or SIGCHLD, or
           for two seconds to elapse (SIGALRM).  pause() should not return. */
        alarm(2);
        pause();

        exit(EXIT_FAILURE);
    }

    /* At this point we are executing as the child process */
    parent = getppid();

    /* Cancel certain signals */
    signal(SIGCHLD,SIG_DFL); /* A child process dies */
    signal(SIGTSTP,SIG_IGN); /* Various TTY signals */
    signal(SIGTTOU,SIG_IGN);
    signal(SIGTTIN,SIG_IGN);
    signal(SIGHUP, SIG_IGN); /* Ignore hangup signal */
    signal(SIGTERM,SIG_DFL); /* Die on SIGTERM */

    /* Change the file mode mask */
    umask(0);

    /* Create a new SID for the child process */
    sid = setsid();
    if (sid < 0) {
        msgOut( LOG_ERR, "unable to create a new session, code %d (%s)\n",
                errno, strerror(errno) );
        exit(EXIT_FAILURE);
    }

    /* Change the current working directory.  This prevents the current
       directory from being locked; hence not being able to remove it. */
    if ((chdir("/")) < 0) {
        msgOut( LOG_ERR, "unable to change directory to %s, code %d (%s)\n",
                "/", errno, strerror(errno) );
        exit(EXIT_FAILURE);
    }

    /* Redirect standard files to /dev/null */
    freopen( "/dev/null", "r", stdin);
    freopen( "/dev/null", "w", stdout);
    freopen( "/dev/null", "w", stderr);

	// Write run ID to lock file
	{
    	char runFile[32];
        sprintf( runFile, "%d\n", getpid() );
        write( lfp, runFile, strlen(runFile) );
	}

    /* Tell the parent process that we are A-okay */
    kill( parent, SIGUSR1 );
}

static int init_udev_monitor_socket(void)
{
	struct sockaddr_un saddr;
	socklen_t addrlen;
	const int feature_on = 1;
	int retval;

	memset(&saddr, 0x00, sizeof(saddr));
	saddr.sun_family = AF_LOCAL;
	/* use abstract namespace for socket path */
	strcpy(&saddr.sun_path[1], "/org/kernel/udev/monitor");
	addrlen = offsetof(struct sockaddr_un, sun_path) + strlen(saddr.sun_path+1) + 1;

	udev_monitor_sock = socket(AF_LOCAL, SOCK_DGRAM, 0);
	if (udev_monitor_sock == -1) {
		msgOut(LOG_ERR, "error getting socket: %s\n", strerror(errno));
		return -1;
	}

	/* the bind takes care of ensuring only one copy running */
	retval = bind(udev_monitor_sock, (struct sockaddr *) &saddr, addrlen);
	if (retval < 0) {
		msgOut(LOG_ERR, "bind failed: %s\n", strerror(errno));
		close(udev_monitor_sock);
		udev_monitor_sock = -1;
		return -1;
	}

	/* enable receiving of the sender credentials */
	setsockopt(udev_monitor_sock, SOL_SOCKET, SO_PASSCRED, &feature_on, sizeof(feature_on));

	return 0;
}

static int init_uevent_netlink_sock(void)
{
	struct sockaddr_nl snl;
	int retval;

	memset(&snl, 0x00, sizeof(struct sockaddr_nl));
	snl.nl_family = AF_NETLINK;
	snl.nl_pid = getpid();
	snl.nl_groups = 1;

	uevent_netlink_sock = socket(PF_NETLINK, SOCK_DGRAM, NETLINK_KOBJECT_UEVENT);
	if (uevent_netlink_sock == -1) {
		msgOut(LOG_ERR, "error getting socket: %s\n", strerror(errno));
		return -1;
	}

	retval = bind(uevent_netlink_sock, (struct sockaddr *) &snl,
		      sizeof(struct sockaddr_nl));
	if (retval < 0) {
		msgOut(LOG_ERR, "bind failed: %s\n", strerror(errno));
		close(uevent_netlink_sock);
		uevent_netlink_sock = -1;
		return -1;
	}

	return 0;
}

// Close handle and mark as closed
void CloseHandle( int& fh )
{
	if (fh >= 0)
	{
		if (!close( fh ))
		{
			// Mark closed iff close() succeeded
			fh = -1;
		}
	}
}

// Close handles we don't want to pass on to children
void CloseCritHandles()
{
	CloseHandle( lfp );
	if (udev)
		CloseHandle( udev_monitor_sock );
	if (kernel)
		CloseHandle( uevent_netlink_sock );
}

// Reopen handles we temporarily closed
int ReopenCritHandles( const char * lockFile )
{
	if (lockFile)
	{
		if (OpenLockfile( lockFile ) < 0)
		{
			udev_exit = 1;
			msgOut( LOG_ERR, "Unable to open lockfile %s\n", lockFile );
			return -1;
		}
	}
	int res = 0;
	if (udev) res += init_udev_monitor_socket();
	if (kernel) res += init_uevent_netlink_sock();
	return res;
}


static void sig_handler(int signum)
{
	if (signum == SIGINT || signum == SIGTERM)
		udev_exit = 1;
}

static const char *search_key(const char *searchkey, const char *buf, size_t buflen)
{
	size_t bufpos = 0;
	size_t searchkeylen = strlen(searchkey);

	while (bufpos < buflen) {
		const char *key;
		int keylen;

		key = &buf[bufpos];
		keylen = strlen(key);
		if (keylen == 0)
			break;
		 if ((strncmp(searchkey, key, searchkeylen) == 0) && key[searchkeylen] == '=')
			return &key[searchkeylen + 1];
		bufpos += keylen + 1;
	}
	return NULL;
}

// Put all key values into environment, clearing previously specified values
void AddKeyValuesToEnv( const char *envStart, int envLength )
{
	int envPos = 0;
	while (envPos < envLength) {
		int keylen;
		const char *key;

		key = &envStart[envPos];
		keylen = strlen(key);
		if (keylen == 0)
			break;
		char *tmp = strdup( key );
		char *name = strtok( tmp, "=" );
		char *val = strtok( NULL, "=" );
		PushEnv( name, val );
		free( tmp );
		//msgOut( LOG_DEBUG, "%s\n", key );
		envPos += (keylen + 1);
	}

}

// Run specified script. Note that scripts may start daemons so we need to close critical
// file handles we don't want passed on
int RunScript( const char *action, const char *script, const char *devpath, int verbose )
{
	int res = -1;
	if (script && *script)
	{
		if (verbose) msgOut( LOG_INFO, "[event %s device %s script %s]\n", action, devpath, script );
		////////// Start critical section - close handles
		if (g_closeHandles) CloseCritHandles();
		FILE *pf = popen( script, "r" );
		if (pf)
		{
			char buff[1024];
			char *mount = NULL;
			char *msg = NULL;
			char *umount = NULL;
			while (fgets( buff, sizeof(buff), pf ))
			{
				int len = strcspn( buff, "\r\n" );
				buff[len] = '\0'; // chomp trailing LF
				if (verbose) msgOut( LOG_INFO, "<%s\n", buff );
				if (buff[0] == '#' || buff[0] == '\0')
				{
					continue;
				}
				if (!strncmp( buff, "RESULT=", 7 ))
				{
					res = atoi( &buff[7] );
				}
				else if (!strncmp( buff, "MOUNT=", 6 ))
				{
					mount = strdup( &buff[6] );
				}
				else if (!strncmp( buff, "UMOUNT=", 7  ))
				{
					umount = strdup( &buff[7] );
				}
				else if (!strncmp( buff, "MESSAGE=", 8 ))
				{
					msg = strdup( &buff[8] );
				}
				// else just ignore it
			}
			if (verbose) msgOut( LOG_INFO, "<EOF>\n" );
			pclose( pf );
			//////////////// End critical section - reopen handles
			if (g_closeHandles)
			{
				if (ReopenCritHandles( LOCKFILE_NAME ) < 0)
				{
					udev_exit = 1;
				}
			}
			if (verbose || (msg != NULL || res != 0)) msgOut( LOG_INFO, "result=%d %s%s%s\n", res, msg ? msg : "", mount ? " mount=" : "", mount ? mount : "" );
			if (umount && *umount)
			{
				ReleaseMountPoint( devpath );
			}
			// Note that a mount returned means we mark it as "engaged" which should become "busy" when the mount
			// event is reported.
			if (mount && *mount)
			{
				if (verbose) msgOut( LOG_INFO, "Marking mount point %s in use by %s\n", mount, devpath );
				// 1 = engaged, 2 = married
				MarkMountPoint( devpath, mount, 1 );
			}
			if (msg) free( msg );
			if (mount) free( mount );
			if (umount) free( umount );
		}
		else
		{
			msgOut( LOG_ERR, "failed to start script!\n" );
		}
	}
	else
	{
		msgOut( LOG_ERR, "[no script defined for %s]\n", action );
	}
	return res;
}

char *helpText = "Syntax: mountmon [options]\n\
	where options are one or more of the following:\n\
	--debug		Increase debug level\n\
	--daemon	Detach and run as a daemon\n\
	--kernel	Show kernel UEVENT as well as UDEV\n\
	--conf=file	Use file for configuration (default=/etc/mountmon/mountmon.conf)\n\
	--help		Display this message\n\
";
int main(int argc, char *argv[])
{
	struct sigaction act;
	int option;
	int env = 0;
	fd_set readfds;
	int retval = 0;
	int nArg;
	MMOptions mm;
	int ruleRes;
	char myPid[32];
	char eventSerial[32];
	char mmVerbose[32];
	char mmDaemon[32];
	int eventSerialNum = 1;
	time_t lastNotice;
	int totalUEvent = 0;
	int totalUDev = 0;

	if (getuid() != 0 && kernel) {
		fprintf(stderr, "root privileges needed to subscribe to kernel events\n");
		goto out;
	}

	for (nArg = 1; nArg < argc; nArg++)
	{
		if (!strcmp( argv[nArg], "-d" ) || !strcmp( argv[nArg], "--debug" ) || !strcmp( argv[nArg], "--verbose" ))
		{
			g_dbg++;
		}
		else if (!strcmp( argv[nArg], "--daemon" ))
		{
			g_daemon = 1;
		}
		else if (!strcmp( argv[nArg], "--env" ))
		{
			env = 1;
		}
		else if (!strcmp( argv[nArg], "--kernel" ))
		{
			kernel = 1;
			udev = 0;
		}
		else if (!strncmp( argv[nArg], "--conf=", 7 ))
		{
			mm.SetConf( &argv[nArg][7] );
		}
		else if (!strcmp( argv[nArg], "--help" ))
		{
			printf( helpText );
			return 0;
		}
		else
		{
			printf( "Error: unrecognized option %s\n", argv[nArg] );
			printf( helpText );
			return -1;
		}
	}

	msgOut( LOG_INFO, "Starting mountmon v" VER_STR "%s\n", g_daemon ? " as daemon" : "" );
	ruleRes = mm.ParseConfig( g_dbg );
	// Allow command line --verbose to take precedence over config file
	if (mm.GetVerbose() >= 0 && g_dbg == 0)
	{
		g_dbg = mm.GetVerbose();
	}
	msgOut( LOG_INFO, "Using configuration file %s (%d rules parsed)\n", mm.GetConf(), ruleRes );
	if (g_daemon)
	{
		openlog( DAEMON_NAME, LOG_PID, LOG_LOCAL5 );
		msgOut( LOG_INFO, "starting\n" );
		daemonize( "/var/run/" DAEMON_NAME ".pid" );
	}

	if (mm.GetEnvCount() > 0)
	{
		msgOut( LOG_INFO, "%d environment values set in config:\n", mm.GetEnvCount() );
		int n;
		for (n = 0; n < mm.GetEnvCount(); n++)
		{
			msgOut( LOG_INFO, "%s=%s\n", mm.GetEnvName(n), mm.GetEnvValue(n) );
			setenv( mm.GetEnvName(n), mm.GetEnvValue(n), 1 );
		}
	}

	msgOut( LOG_INFO, "Checking existing mounts: %d found\n", CheckExistingMounts() );
	DumpMountPoints( 2 );

	/* set signal handlers */
	memset(&act, 0x00, sizeof(struct sigaction));
	act.sa_handler = (void (*)(int)) sig_handler;
	sigemptyset(&act.sa_mask);
	act.sa_flags = SA_RESTART;
	sigaction(SIGINT, &act, NULL);
	sigaction(SIGTERM, &act, NULL);

	msgOut( LOG_INFO, "mountmon will print the received events for:\n");
	if (ReopenCritHandles( NULL ) < 0)
	{
		goto out;
	}
	if (udev) {
		msgOut( LOG_INFO, "UDEV the event which udev sends out after rule processing\n");
	}
	if (kernel) {
		msgOut( LOG_INFO, "UEVENT the kernel uevent\n");
	}

	// Set static environment values
	sprintf( myPid, "%d", getpid() );
	setenv( "MM_PID", myPid, 1 );
	setenv( "MM_VER", VER_STR, 1 );
	sprintf( mmVerbose, "%d", g_dbg );
	sprintf( mmDaemon, "%d", g_daemon );
	setenv( "MM_VERBOSE", mmVerbose, 1 );
	setenv( "MM_DAEMON", mmDaemon, 1 );

	msgOut( LOG_INFO, "Starting monitor... PATH=%s\n", getenv("PATH") );
	lastNotice = time(NULL);
	while (!udev_exit) {
		char buf[UEVENT_BUFFER_SIZE*2];
		ssize_t buflen;
		ssize_t bufpos;
		ssize_t keys;
		int fdcount;
		struct timeval tv;
		struct timezone tz;
		char timestr[64];
		const char *source = NULL;
		const char *devpath, *action, *subsys;
		const char *vendorName, *modelName, *devName;
		int isUdev = 0;
		time_t now = time(NULL);

		// Say something every ten minutes
		if (now - lastNotice > 600)
		{
			msgOut( LOG_INFO, "mountmon: %d udev %d uevent\n", totalUDev, totalUEvent );
		}

		buflen = 0;
		FD_ZERO(&readfds);
		if (uevent_netlink_sock >= 0)
			FD_SET(uevent_netlink_sock, &readfds);
		if (udev_monitor_sock >= 0)
			FD_SET(udev_monitor_sock, &readfds);

		fdcount = select(UDEV_MAX(uevent_netlink_sock, udev_monitor_sock)+1, &readfds, NULL, NULL, NULL);
		if (fdcount < 0) {
			if (errno != EINTR)
				msgOut(LOG_ERR, "error receiving uevent message: %s\n", strerror(errno));
			continue;
		}

		if (gettimeofday(&tv, &tz) == 0) {
			snprintf(timestr, sizeof(timestr), "%llu.%06u",
				 (unsigned long long) tv.tv_sec, (unsigned int) tv.tv_usec);
		} else
			timestr[0] = '\0';

		if ((uevent_netlink_sock >= 0) && FD_ISSET(uevent_netlink_sock, &readfds)) {
			buflen = recv(uevent_netlink_sock, &buf, sizeof(buf), 0);
			if (buflen <= 0) {
				msgOut(LOG_ERR, "error receiving uevent message: %s\n", strerror(errno));
				continue;
			}
			totalUEvent++;
			source = "UEVENT";
		}

		if ((udev_monitor_sock >= 0) && FD_ISSET(udev_monitor_sock, &readfds)) {
			buflen = recv(udev_monitor_sock, &buf, sizeof(buf), 0);
			if (buflen <= 0) {
				msgOut(LOG_ERR, "error receiving udev message: %s\n", strerror(errno));
				continue;
			}
			totalUDev++;
			isUdev = 1;
			source = "UDEV  ";
		}

		if (buflen == 0)
			continue;

		keys = strlen(buf) + 1; /* start of payload */

		// Clear all previously set environment values
		ReleaseEnv();

		// Add all values to environment
		AddKeyValuesToEnv( &buf[keys], buflen - keys );

		devpath = search_key("DEVPATH", &buf[keys], buflen);
		action = search_key("ACTION", &buf[keys], buflen);
		subsys = search_key("SUBSYSTEM", &buf[keys], buflen);
		vendorName = search_key("ID_VENDOR", &buf[keys], buflen);
		modelName = search_key("ID_MODEL", &buf[keys], buflen);
		devName = search_key("DEVNAME", &buf[keys], buflen);
		if (g_dbg) msgOut( LOG_DEBUG, "%s[%s] %-8s %s (%s) vendor=%s model=%s devnode=%s\n", source, timestr, action, devpath, subsys,
			vendorName, modelName, devName );

		// Additional values. Some static values are set at startup time
		PushEnv( "MM_UDEV", isUdev ? "1" : "0" );
		PushEnv( "MM_UEVENT", isUdev ? "0" : "1" );
		sprintf( eventSerial, "%d", eventSerialNum );
		eventSerialNum++;
		PushEnv( "MM_SERIAL", eventSerial );
		PushEnv( "MM_PREVIOUS_MOUNTPOINT", GetMountHistory( devpath ) );
		PushEnv( "MM_CURRENT_MOUNTPOINT", GetCurrentMountPoint( devpath ) );

		// Actions we handle:
		// add Apple/iPod - determine filesystem type, mount
		// remove Apple/iPod - unmount
		// mount - mark entry as in use
		// umount - free available mount point
		if (isUdev && !strcmp( action, "add" ) /*&& vendorName && modelName && devName*/)
		{
			RunScript( action, mm.GetAddCmd(), devpath, g_dbg );
		}
		if (isUdev && !strcmp( action, "remove" ) && devpath /* && vendorName && modelName && devName*/)
		{
			RunScript( action, mm.GetRemoveCmd(), devpath, g_dbg );
			ReleaseMountPoint( devpath );
		}
		if (isUdev && !strcmp( action, "mount" ))
		{
			RunScript( action, mm.GetMountCmd(), devpath, g_dbg );
			// Change provisional status to confirmed mount status
			MarkMountPoint( devpath, GetCurrentMountPoint( devpath ), 2 );
			if (g_dbg) DumpMountPoints( 0 );
		}
		if (isUdev && !strcmp( action, "umount" ))
		{
			RunScript( action, mm.GetUmountCmd(), devpath, g_dbg );
			ReleaseMountPoint( devpath );
			if (g_dbg) DumpMountPoints( 0 );
		}

		/* print environment */
		bufpos = keys;
		if (env) {
			while (bufpos < buflen) {
				int keylen;
				char *key;

				key = &buf[bufpos];
				keylen = strlen(key);
				if (keylen == 0)
					break;
				msgOut( LOG_DEBUG, "%s\n", key );
				bufpos += keylen + 1;
			}
		}
	}

	msgOut( LOG_INFO, "Exit requested\n" );
out:
	msgOut( LOG_INFO, "Exiting... return value = %d\n", retval );

	CloseCritHandles();

	if (g_daemon)
	{
		/* Finish up */
		msgOut( LOG_NOTICE, "terminated\n" );
		closelog();
	}

	if (retval)
		return 1;
	return 0;
}
